from .. import api

import pandas as pd
from pandas.api.types import is_numeric_dtype

description = """
Summarize a table file.
"""

epilog = f"""
[Example] Summarize a table:
  $ fuc {api.common._script_name()} table.tsv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Summarize a table file.',
    )
    parser.add_argument(
        'table_file',
        help='Table file.'
    )
    parser.add_argument(
        '--sep',
        metavar='TEXT',
        default='\t',
        help="Delimiter to use (default: '\\t')."
    )
    parser.add_argument(
        '--skiprows',
        metavar='TEXT',
        help='Comma-separated line numbers to skip (0-indexed) or \n'
             'number of lines to skip at the start of the file \n'
             '(e.g. `--skiprows 1,` will skip the second line, \n'
             '`--skiprows 2,4` will skip the third and fifth lines, \n'
             'and `--skiprows 10` will skip the first 10 lines).'
    )
    parser.add_argument(
        '--na_values',
        metavar='TEXT',
        nargs='+',
        help='Additional strings to recognize as NA/NaN (by \n'
             "default, the following values are interpreted \n"
             "as NaN: '', '#N/A', '#N/A N/A', '#NA', '-1.#IND', \n"
             "'-1.#QNAN', '-NaN', '-nan', '1.#IND', '1.#QNAN', \n"
             "'<NA>', 'N/A', 'NA', 'NULL', 'NaN', 'n/a', 'nan', \n"
             "'null')."
    )
    parser.add_argument(
        '--keep_default_na',
        action='store_false',
        help='Whether or not to include the default NaN values when \n'
             "parsing the data (see 'pandas.read_table' for details)."
    )
    parser.add_argument(
        '--expr',
        metavar='TEXT',
        help='Query the columns of a pandas.DataFrame with a \n'
             """boolean expression (e.g. `--query "A == 'yes'"`)."""
    )
    parser.add_argument(
        '--columns',
        metavar='TEXT',
        nargs='+',
        help='Columns to be summarized (by default, all columns \n'
             'will be included).'
    )
    parser.add_argument(
        '--dtypes',
        metavar='PATH',
        help="File of column names and their data types (either \n"
             "'categorical' or 'numeric'); one tab-delimited pair of \n"
             "column name and data type per line."
    )

def main(args):
    if args.skiprows is None:
        skiprows = None
    elif ',' in args.skiprows:
        skiprows = [int(x) for x in args.skiprows.split(',') if x]
    else:
        skiprows = int(args.skiprows)

    df = pd.read_table(args.table_file,
                       sep=args.sep,
                       skiprows=skiprows,
                       na_values=args.na_values,
                       keep_default_na=args.keep_default_na)

    if args.dtypes is not None:
        with open(args.dtypes) as f:
            for line in f:
                fields = line.strip().split('\t')
                col = fields[0]
                dtype = fields[1]
                if dtype == 'numeric':
                    dtype = float
                elif dtype == 'categorical':
                    dtype = 'category'
                else:
                    raise TypeError("Data type must be either 'numeric' or "
                        "'categorical'.")
                df[col] = df[col].astype(dtype)

    if args.expr:
        df = df.query(args.expr)

    if args.columns:
        headers = args.columns
    else:
        headers = df.columns

    print('='*70)

    for header in headers:
        s = df[header]
        print(f'Name: {header}')
        print(f'Count: {len(s)}')
        print('Summary:')
        if is_numeric_dtype(s):
            print(s.describe().to_string())
        else:
            print(s.value_counts().sort_index().to_string())
        print('='*70)



import sys

from .. import api

import pysam

description = """
Concatenate depth of coverage files.
"""

epilog = f"""
[Example] Concatenate vertically:
  $ fuc {api.common._script_name()} in1.tsv in2.tsv > out.tsv

[Example] Concatenate horizontally:
  $ fuc {api.common._script_name()} in1.tsv in2.tsv --axis 1 > out.tsv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Concatenate depth of coverage files.',
    )
    parser.add_argument(
        'tsv',
        metavar='PATH',
        nargs='+',
        help='One or more TSV files.'
    )
    parser.add_argument(
        '--axis',
        metavar='INT',
        default=0,
        type=int,
        help='The axis to concatenate along (default: 0) (choices: \n'
             '0, 1 where 0 is index and 1 is columns).'
    )

def main(args):
    cfs = [api.pycov.CovFrame.from_file(x) for x in args.tsv]
    cf = api.pycov.concat(cfs, axis=args.axis)
    sys.stdout.write(cf.to_string())



from setuptools import setup, find_packages

exec(open('fuc/version.py').read())

requirements = [
    'biopython', 'lxml', 'matplotlib', 'matplotlib-venn', 'numpy', 'pandas',
    'pyranges', 'pysam', 'scipy', 'seaborn', 'statsmodels'
]

setup(
    name='fuc',
    version=__version__,
    author='Seung-been "Steven" Lee',
    author_email='sbstevenlee@gmail.com',
    description='Frequently used commands in bioinformatics',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/sbslee/fuc',
    packages=find_packages(),
    license='MIT',
    install_requires=requirements,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License'
    ],
    entry_points={'console_scripts': ['fuc=fuc.__main__:main']}
)



import sys

from .. import api

import pysam
from fuc import common, pyvcf

description = """
Slice a GFF/BED/SAM/VCF file with Tabix.

After creating an index file (.tbi), the Tabix program is able to quickly
retrieve data lines overlapping regions specified in the format
'chr:start-end'. Coordinates specified in this region format are 1-based and
inclusive.
"""

epilog = f"""
[Example] Slice a VCF file:
  $ fuc {api.common._script_name()} in.vcf.gz chr1:100-200 > out.vcf
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Slice a GFF/BED/SAM/VCF file with Tabix.',
    )
    parser.add_argument(
        'file',
        help='File to be sliced.'
    )
    parser.add_argument(
        'regions',
        nargs='+',
        help='One or more regions.'
    )

def main(args):
    if '.vcf' in args.file:
        vf = pyvcf.VcfFrame.from_file(args.file, meta_only=True)
        for line in vf.meta:
            sys.stdout.write(line + '\n')
        sys.stdout.write('#' + '\t'.join(vf.df.columns) + '\n')
    elif '.sam' in args.file:
        pass
    elif '.gff' in args.file:
        pass
    elif '.bed' in args.file:
        pass
    else:
        raise ValueError('Unsupported file format')

    tbx = pysam.TabixFile(args.file)

    for region in args.regions:
        for row in tbx.fetch(*common.parse_region(region)):
             sys.stdout.write(row + '\n')



import sys
from pathlib import Path

from .. import api

description = """
Check whether certain files exist.

This command will check whether or not specified files including directories
exist, returning 'True' if they exist and 'False' otherwise.
"""

epilog = f"""
[Example] Test a file:
  $ fuc {api.common._script_name()} in.txt

[Example] Test a directory:
  $ fuc {api.common._script_name()} dir

[Example] When the input is stdin:
  $ cat test.list | fuc {api.common._script_name()}
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Check whether certain files exist.',
    )
    parser.add_argument('files', nargs='*',
        help='Files and directories to be tested (default: stdin).')

def main(args):
    if args.files:
        paths = args.files
    elif not sys.stdin.isatty():
        paths = sys.stdin.read().rstrip('\n').split('\n')
    else:
        raise ValueError('no input files detected')
    for path in paths:
        print(Path(path).exists(), path)



"""
The pybed submodule is designed for working with BED files. It
implements ``pybed.BedFrame`` which stores BED data as ``pandas.DataFrame``
via the `pyranges <https://github.com/biocore-ntnu/pyranges>`_ package to
allow fast computation and easy manipulation. The submodule strictly adheres
to the standard `BED specification
<https://genome.ucsc.edu/FAQ/FAQformat.html>`_.

BED lines can have the following fields (the first three are required):

+-----+-------------+-----------------------------------+--------------+
| No. | Name        | Description                       | Examples     |
+=====+=============+===================================+==============+
| 1   | Chromosome  | Chromosome                        | 'chr2', '2'  |
+-----+-------------+-----------------------------------+--------------+
| 2   | Start       | Start position                    | 10041, 23042 |
+-----+-------------+-----------------------------------+--------------+
| 3   | End         | End position                      | 10041, 23042 |
+-----+-------------+-----------------------------------+--------------+
| 4   | Name        | Feature name                      | 'TP53'       |
+-----+-------------+-----------------------------------+--------------+
| 5   | Score       | Score for color density (0, 1000) | 342, 544     |
+-----+-------------+-----------------------------------+--------------+
| 6   | Strand      | '+' or '-' ('.' for no strand)    | '+', '-'     |
+-----+-------------+-----------------------------------+--------------+
| 7   | ThickStart  | Start position for thick drawing  | 10041, 23042 |
+-----+-------------+-----------------------------------+--------------+
| 8   | ThickEnd    | End position for thick drawing    | 10041, 23042 |
+-----+-------------+-----------------------------------+--------------+
| 9   | ItemRGB     | RGB value                         | '255,0,0'    |
+-----+-------------+-----------------------------------+--------------+
| 10  | BlockCount  | Number of blocks (e.g. exons)     | 12, 8        |
+-----+-------------+-----------------------------------+--------------+
| 11  | BlockSizes  | ','-separated block sizes         | '224,423'    |
+-----+-------------+-----------------------------------+--------------+
| 12  | BlockStarts | ','-separated block starts        | '2345,5245'  |
+-----+-------------+-----------------------------------+--------------+
"""

import pandas as pd
import pyranges as pr
from copy import deepcopy
from . import common

HEADERS = [
    'Chromosome', 'Start', 'End', 'Name', 'Score', 'Strand', 'ThickStart',
    'ThickEnd', 'ItemRGB', 'BlockCount', 'BlockSizes', 'BlockStarts'
]

class BedFrame:
    """Class for storing BED data.

    Parameters
    ----------
    meta : list
        Metadata lines.
    gr : pyranges.PyRanges
        PyRanges object containing BED data.

    See Also
    --------
    BedFrame.from_dict
        Construct BedFrame from a dict of array-like or dicts.
    BedFrame.from_file
        Construct BedFrame from a BED file.
    BedFrame.from_frame
        Construct BedFrame from a dataframe.
    BedFrame.from_region
        Construct BedFrame from a list of regions.

    Examples
    --------

    >>> import pandas as pd
    >>> import pyranges as pr
    >>> from fuc import pybed
    >>> data = {
    ...     'Chromosome': ['chr1', 'chr2', 'chr3'],
    ...     'Start': [100, 400, 100],
    ...     'End': [200, 500, 200]
    ... }
    >>> df = pd.DataFrame(data)
    >>> gr = pr.PyRanges(df)
    >>> bf = pybed.BedFrame([], gr)
    >>> bf.gr.df
      Chromosome  Start  End
    0       chr1    100  200
    1       chr2    400  500
    2       chr3    100  200
    """
    def __init__(self, meta, gr):
        self._meta = meta
        self._gr = gr

    @property
    def meta(self):
        """list : Metadata lines."""
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value

    @property
    def gr(self):
        """pyranges.PyRanges : Two-dimensional representation of genomic
        intervals and their annotations."""
        return self._gr

    @gr.setter
    def gr(self, value):
        self._gr = value

    @property
    def contigs(self):
        """list : List of contig names."""
        return self.gr.chromosomes

    @property
    def shape(self):
        """tuple : Dimensionality of BedFrame (intervals, columns)."""
        return self.gr.df.shape

    @property
    def has_chr_prefix(self):
        """bool : Whether the (annoying) 'chr' string is found."""
        for contig in self.contigs:
            if 'chr' in contig:
                return True
        return False

    def copy_meta(self):
        """Return a copy of the metadata."""
        return deepcopy(self.meta)

    def to_file(self, fn):
        """Write the BedFrame to a BED file."""
        with open(fn, 'w') as f:
            if self.meta:
                f.write('\n'.join(self.meta) + '\n')
            self.gr.df.to_csv(f, sep='\t', index=False, header=False)

    def to_string(self):
        """Render the BedFrame to a console-friendly tabular output."""
        s = ''
        if self.meta:
            s += '\n'.join(self.meta) + '\n'
        s += self.gr.df.to_csv(header=False, index=False, sep='\t')
        return s

    def intersect(self, other):
        """Find intersection between the BedFrames."""
        bf = self.__class__(deepcopy(self.meta), self.gr.intersect(other.gr))
        return bf

    @classmethod
    def from_dict(cls, meta, data):
        """
        Construct BedFrame from a dict of array-like or dicts.

        Parameters
        ----------
        meta : list
            Metadata lines.
        data : dict
            Of the form {field : array-like} or {field : dict}.

        Returns
        -------
        BedFrame
            BedFrame object.

        See Also
        --------
        BedFrame
            BedFrame object creation using constructor.
        BedFrame.from_file
            Construct BedFrame from a BED file.
        BedFrame.from_frame
            Construct BedFrame from a dataframe.
        BedFrame.from_region
            Construct BedFrame from a list of regions.

        Examples
        --------

        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr2', 'chr3'],
        ...     'Start': [100, 400, 100],
        ...     'End': [200, 500, 200]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.gr.df
          Chromosome  Start  End
        0       chr1    100  200
        1       chr2    400  500
        2       chr3    100  200
        """
        return cls(meta, pr.PyRanges(pd.DataFrame(data)))

    @classmethod
    def from_file(cls, fn):
        """
        Construct BedFrame from a BED file.

        Parameters
        ----------
        fn : str
            BED file path.

        Returns
        -------
        BedFrame
            BedFrame object.

        See Also
        --------
        BedFrame
            BedFrame object creation using constructor.
        BedFrame.from_dict
            Construct BedFrame from a dict of array-like or dicts.
        BedFrame.from_frame
            Construct BedFrame from a dataframe.
        BedFrame.from_region
            Construct BedFrame from a list of regions.

        Examples
        --------

        >>> from fuc import pybed
        >>> bf = pybed.BedFrame.from_file('example.bed')
        """
        meta = []
        skip_rows = 0
        with open(fn, 'r') as f:
            for line in f:
                if 'browser' in line or 'track' in line:
                    meta.append(line.strip())
                    skip_rows += 1
                else:
                    headers = HEADERS[:len(line.strip().split())]
                    break
        df = pd.read_table(fn, header=None, names=headers, skiprows=skip_rows)
        return cls(meta, pr.PyRanges(df))

    @classmethod
    def from_frame(cls, meta, data):
        """
        Construct BedFrame from a dataframe.

        Parameters
        ----------
        meta : list
            Metadata lines.
        data : pandas.DataFrame
            DataFrame containing BED data.

        Returns
        -------
        BedFrame
            BedFrame object.

        See Also
        --------
        BedFrame
            BedFrame object creation using constructor.
        BedFrame.from_dict
            Construct BedFrame from a dict of array-like or dicts.
        BedFrame.from_file
            Construct BedFrame from a BED file.
        BedFrame.from_region
            Construct BedFrame from a list of regions.

        Examples
        --------

        >>> import pandas as pd
        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr2', 'chr3'],
        ...     'Start': [100, 400, 100],
        ...     'End': [200, 500, 200]
        ... }
        >>> df = pd.DataFrame(data)
        >>> bf = pybed.BedFrame.from_frame([], df)
        >>> bf.gr.df
          Chromosome  Start  End
        0       chr1    100  200
        1       chr2    400  500
        2       chr3    100  200
        """
        return cls(meta, pr.PyRanges(data))

    @classmethod
    def from_regions(cls, meta, regions):
        """
        Construct BedFrame from a list of regions.

        Parameters
        ----------
        meta : list
            Metadata lines.
        regions : str or list
            Region or list of regions.

        Returns
        -------
        BedFrame
            BedFrame object.

        See Also
        --------
        BedFrame
            BedFrame object creation using constructor.
        BedFrame.from_dict
            Construct BedFrame from a dict of array-like or dicts.
        BedFrame.from_file
            Construct BedFrame from a BED file.
        BedFrame.from_frame
            Construct BedFrame from a dataframe.

        Examples
        --------

        >>> from fuc import pybed
        >>> data = ['chr1:100-200', 'chr2:100-200', 'chr3:100-200']
        >>> bf = pybed.BedFrame.from_regions([], data)
        >>> bf.gr.df
          Chromosome  Start  End
        0       chr1    100  200
        1       chr2    100  200
        2       chr3    100  200
        """
        if isinstance(regions, str):
            regions = [regions]
        chroms = []
        starts = []
        ends = []
        for region in regions:
            chrom, start, end = common.parse_region(region)
            chroms.append(chrom)
            starts.append(start)
            ends.append(end)
        data = {'Chromosome': chroms, 'Start': starts, 'End': ends}
        return cls.from_dict([], data)

    def update_chr_prefix(self, mode='remove'):
        """
        Add or remove the (annoying) 'chr' string from the Chromosome column.

        Parameters
        ----------
        mode : {'add', 'remove'}, default: 'remove'
            Whether to add or remove the 'chr' string.

        Returns
        -------
        BedFrame
            Updated BedFrame.

        Examples
        --------

        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['1', '1', 'chr2', 'chr2'],
        ...     'Start': [100, 400, 100, 200],
        ...     'End': [200, 500, 200, 300]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.gr.df
          Chromosome  Start  End
        0          1    100  200
        1          1    400  500
        2       chr2    100  200
        3       chr2    200  300
        >>> bf.update_chr_prefix(mode='remove').gr.df
          Chromosome  Start  End
        0          1    100  200
        1          1    400  500
        2          2    100  200
        3          2    200  300
        >>> bf.update_chr_prefix(mode='add').gr.df
          Chromosome  Start  End
        0       chr1    100  200
        1       chr1    400  500
        2       chr2    100  200
        3       chr2    200  300
        """
        if mode == 'remove':
            def one_row(r):
                r.Chromosome = r.Chromosome.replace('chr', '')
                return r
        elif mode == 'add':
            def one_row(r):
                if 'chr' not in r.Chromosome:
                    r.Chromosome = 'chr' + r.Chromosome
                return r
        else:
            raise ValueError(f'Incorrect mode: {mode}')
        df = self.gr.df.apply(one_row, axis=1)
        return self.__class__([], pr.PyRanges(df))

    def sort(self):
        """
        Sort the BedFrame by chromosome and position.

        Returns
        -------
        BedFrame
            Sorted BedFrame.

        Examples
        --------

        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr3', 'chr1'],
        ...     'Start': [400, 100, 100],
        ...     'End': [500, 200, 200]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.gr.df
          Chromosome  Start  End
        0       chr1    400  500
        1       chr1    100  200
        2       chr3    100  200
        >>> bf.sort().gr.df
          Chromosome  Start  End
        0       chr1    100  200
        1       chr1    400  500
        2       chr3    100  200
        """
        return self.__class__(self.copy_meta(), self.gr.sort())

    def merge(self):
        """
        Merge overlapping intervals within BedFrame.

        Returns
        -------
        BedFrame
            Merged BedFrame.

        Examples
        --------

        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr1', 'chr2', 'chr2', 'chr3', 'chr3'],
        ...     'Start': [10, 30, 15, 25, 50, 61],
        ...     'End': [40, 50, 25, 35, 60, 80]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.gr.df
          Chromosome  Start  End
        0       chr1     10   40
        1       chr1     30   50
        2       chr2     15   25
        3       chr2     25   35
        4       chr3     50   60
        5       chr3     61   80
        >>> bf.merge().gr.df
          Chromosome  Start  End
        0       chr1     10   50
        1       chr2     15   35
        2       chr3     50   60
        3       chr3     61   80
        """
        return self.__class__(self.copy_meta(), self.gr.merge())

    def to_regions(self, merge=True):
        """
        Return a list of regions from BedFrame.

        Parameters
        ----------
        merge : bool, default: True
            Whether to merge overlapping intervals.

        Returns
        -------
        list
            List of regions.

        Examples
        --------
        >>> from fuc import pybed
        >>> data = {
        ...     'Chromosome': ['chr1', 'chr1', 'chr2', 'chr2', 'chr3', 'chr3'],
        ...     'Start': [10, 30, 15, 25, 50, 61],
        ...     'End': [40, 50, 25, 35, 60, 80]
        ... }
        >>> bf = pybed.BedFrame.from_dict([], data)
        >>> bf.regions()
        ['chr1:10-50', 'chr2:15-35', 'chr3:50-60', 'chr3:61-80']
        >>> bf.regions(merge=False)
        ['chr1:10-40', 'chr1:30-50', 'chr2:15-25', 'chr2:25-35', 'chr3:50-60', 'chr3:61-80']
        """
        if merge:
            bf = self.merge()
        else:
            bf = self
        s = bf.gr.df.apply(lambda r: f'{r.Chromosome}:{r.Start}-{r.End}', axis=1)
        return s.to_list()



import sys

from .. import api

import pandas as pd

description = """
Rename the samples in a depth of coverage file.

There are three different renaming modes using the names file:
  - 'MAP': Default mode. Requires two columns, old names in the first
    and new names in the second.
  - 'INDEX': Requires two columns, new names in the first and 0-based
    indicies in the second.
  - 'RANGE': Requires only one column of new names but --range must
    be specified.
"""

epilog = f"""
[Example] Using the default 'MAP' mode:
  $ fuc {api.common._script_name()} in.tsv old_new.tsv > out.tsv

[Example] Using the 'INDEX' mode:
  $ fuc {api.common._script_name()} in.tsv new_idx.tsv --mode INDEX > out.tsv

[Example] Using the 'RANGE' mode:
  $ fuc {api.common._script_name()} in.tsv new_only.tsv --mode RANGE --range 2 5 > out.tsv
"""

def create_parser(subparsers):
    parser = api.common._add_parser(
        subparsers,
        api.common._script_name(),
        description=description,
        epilog=epilog,
        help='Rename the samples in a depth of coverage file.',
    )
    parser.add_argument(
        'tsv',
        help='TSV file (zipped or unzipped).'
    )
    parser.add_argument(
        'names',
        help='Text file containing information for renaming the samples.'
    )
    parser.add_argument(
        '--mode',
        metavar='TEXT',
        default='MAP',
        choices=['MAP', 'INDEX', 'RANGE'],
        help="Renaming mode (default: 'MAP') (choices: 'MAP', \n"
             "'INDEX', 'RANGE')."
    )
    parser.add_argument(
        '--range',
        metavar='INT',
        type=int,
        nargs=2,
        help="Index range to use when renaming the samples. \n"
             "Applicable only with the 'RANGE' mode."
    )
    parser.add_argument(
        '--sep',
        metavar='TEXT',
        default='\t',
        help="Delimiter to use when reading the names file \n"
             "(default: '\\t')."
    )

def main(args):
    cf = api.pycov.CovFrame.from_file(args.tsv)
    df = pd.read_table(args.names, sep=args.sep, header=None)
    if args.mode == 'MAP':
        names = dict(zip(df[0], df[1]))
        indicies = None
    elif args.mode == 'INDEX':
        names = df[0].to_list()
        indicies = df[1].to_list()
    else:
        names = df[0].to_list()
        indicies = tuple(args.range)
    cf = cf.rename(names, indicies=indicies)
    sys.stdout.write(cf.to_string())
